//
// SydB☮x: seccomp and landlock based application sandbox with support for namespaces
// src/test/main.rs: Run integration tests with TAP output
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    env,
    ffi::CString,
    ops::Range,
    path::{Path, PathBuf},
    process::ExitCode,
};

use anyhow::{Context, Result};
use nix::{
    errno::Errno,
    libc::mkdtemp,
    unistd::{chdir, getuid, Uid},
};
use regex::Regex;

mod test;
mod util;
use test::*;

#[derive(Debug)]
struct TempDir {
    path: Option<PathBuf>,
}

impl TempDir {
    fn new<P: AsRef<Path>>(path: P) -> Self {
        TempDir {
            path: Some(path.as_ref().to_path_buf()),
        }
    }
}

impl Drop for TempDir {
    fn drop(&mut self) {
        if let Some(ref path) = self.path {
            if let Err(error) = std::fs::remove_dir_all(path) {
                let path = path.display();
                eprintln!("Failed to remove temporary directory \"{path}\": {error}.");
            }
        }
    }
}

#[derive(Debug)]
enum Arguments {
    Index(usize),
    Range(Range<usize>),
    Pattern(String),
}

struct ArgVec(Vec<Arguments>);

impl From<String> for Arguments {
    fn from(arg: String) -> Self {
        if let Ok(idx) = arg.parse::<usize>() {
            Arguments::Index(idx)
        } else if let Some(range) = arg.split_once("..") {
            if let (Ok(start), Ok(end)) = (range.0.parse::<usize>(), range.1.parse::<usize>()) {
                Arguments::Range(start..end)
            } else {
                Arguments::Pattern(arg)
            }
        } else {
            Arguments::Pattern(arg)
        }
    }
}

impl From<String> for ArgVec {
    fn from(arg: String) -> Self {
        if let Ok(idx) = arg.parse::<usize>() {
            ArgVec(vec![Arguments::Index(idx)])
        } else if let Some(range) = arg.split_once("..") {
            if let (Ok(start), Ok(end)) = (range.0.parse::<usize>(), range.1.parse::<usize>()) {
                ArgVec(vec![Arguments::Range(start..end)])
            } else {
                ArgVec(vec![Arguments::Pattern(arg)])
            }
        } else {
            ArgVec(vec![Arguments::Pattern(arg)])
        }
    }
}

fn main() -> Result<ExitCode> {
    println!("# syd-test: Welcome to the Machine!");
    println!("# usage: syd -t, --test [<name-regex>|<number>|<number>..<number>]..");
    std::env::set_var("SYD_NO_SYSLOG", "YesPlease");

    // Refuse to run the tests as root.
    // The tests will randomly fail when run with elevated privileges.
    if getuid() == Uid::from_raw(0) {
        println!("# running tests as root.");
        println!("# cowardly refusing to continue!");
        println!("1..0 # SKIP running as root.");
        return Ok(ExitCode::from(1));
    }

    // Create a temporary directory and enter it, failures are OK.
    // The directory is removed when the guard is dropped.
    let _guard = {
        #[allow(clippy::disallowed_methods)]
        let mut tmp = CString::new("/tmp/syd_test_XXXXXX")
            .unwrap()
            .into_bytes_with_nul();
        let ptr = unsafe { mkdtemp(tmp.as_mut_ptr() as *mut _) };
        if !ptr.is_null() {
            let path = unsafe { std::ffi::CStr::from_ptr(ptr).to_str()? };
            match chdir(path) {
                Ok(_) => {
                    println!("# running tests under '{path}'.");
                    Some(TempDir::new(path))
                }
                Err(error) => {
                    println!("# chdir failed: {error}.");
                    None
                }
            }
        } else {
            println!("# mkdtemp failed: {:?}.", Errno::last());
            None
        }
    };

    let mut test_indices = Vec::new();

    // Step 1: Handle the SYD_TEST environment variable.
    let mut test_env_arg = false;
    if let Ok(env) = std::env::var("SYD_TEST") {
        if !env.is_empty() {
            test_env_arg = true;

            let arg: Arguments = env.into();
            match arg {
                Arguments::Index(i) => test_indices.push(i),
                Arguments::Range(r) => test_indices.extend(r),
                Arguments::Pattern(p) => {
                    let regex = Regex::new(&p).context("Bad regex pattern")?;
                    for (idx, (name, _)) in TESTS.iter().enumerate() {
                        if regex.is_match(name) {
                            test_indices.push(idx + 1);
                        }
                    }
                }
            }
        }
    }

    // Step 2: Handle command line arguments.
    let args = std::env::args().skip(1).collect::<Vec<_>>();
    let args_is_empty = args.is_empty();
    let args: Vec<Arguments> = args
        .into_iter()
        .map(ArgVec::from)
        .flat_map(|arg_vec| arg_vec.0)
        .collect();
    for arg in args {
        match arg {
            Arguments::Index(i) => test_indices.push(i),
            Arguments::Range(r) => test_indices.extend(r),
            Arguments::Pattern(p) => {
                let regex = Regex::new(&p).context("Bad regex pattern")?;
                for (idx, (name, _)) in TESTS.iter().enumerate() {
                    if regex.is_match(name) {
                        test_indices.push(idx + 1);
                    }
                }
            }
        }
    }

    // If SYD_TEST was not set and no arguments are provided,
    // run all tests.
    if !test_env_arg && args_is_empty {
        test_indices.extend(1..=TEST_COUNT);
    }

    // Print TAP plan.
    if syd::syd_enabled() {
        println!("1..0 # SKIP SydB☮x is enabled!");
        return Ok(ExitCode::from(0));
    }
    println!("1..{TEST_COUNT}");

    let exp_test = env::var("SYD_TEST_EXPENSIVE").ok().is_some();
    let mut skip = 0;
    let mut fail = 0;
    for (idx, (name, test)) in TESTS.iter().enumerate().take(TEST_COUNT) {
        if test_indices.contains(&(idx + 1)) {
            if !exp_test && name.contains("fuzz") {
                println!(
                    "# ok {} - {} # SKIP expensive test, set SYD_TEST_EXPENSIVE to run",
                    idx + 1,
                    name
                );
                continue;
            }
            match test() {
                Ok(_) => {
                    println!("ok {} - {}", idx + 1, name);
                }
                Err(error) => {
                    println!("not ok {} - {} - FAIL: {error}", idx + 1, name);
                    fail += 1;
                }
            }
        } else {
            println!("ok {} - {} # SKIP skipped by command line", idx + 1, name);
            skip += 1;
        }
    }

    let succ = TEST_COUNT - fail - skip;
    println!("# {succ} tests passed.");
    println!("# {fail} tests failed.");
    println!("# {skip} tests skipped.");
    Ok(ExitCode::from(fail.try_into().unwrap_or(127)))
}
