//
// SydB☮x: seccomp and landlock based application sandbox with support for namespaces
// src/test/util.rs: Utilities for integration tests
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

#![allow(clippy::disallowed_methods)]

use std::{env, error::Error, fmt, fs::canonicalize, process::Command, time::Duration};

use nix::errno::Errno;
use once_cell::sync::Lazy;

#[derive(Debug)]
pub struct TestError(pub String);
pub type TestResult = Result<(), TestError>;

impl fmt::Display for TestError {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{}", self.0)
    }
}

impl<E: Error> From<E> for TestError {
    fn from(err: E) -> Self {
        TestError(err.to_string())
    }
}

#[macro_export]
macro_rules! assert {
    ($cond:expr) => {
        if !$cond {
            return Err(TestError(format!("Assertion failed: {}", stringify!($cond))));
        }
    };
    ($cond:expr, $($arg:tt)*) => {
        if !$cond {
            return Err(TestError(format!("Assertion failed: {}: {}", stringify!($cond), format_args!($($arg)*))));
        }
    };
}

#[macro_export]
macro_rules! assert_eq {
    ($left:expr, $right:expr) => {
        if $left != $right {
            return Err(TestError(format!("Assertion failed in {} ({}:{}): (left: `{}`, right: `{}`)", std::any::type_name::<fn()>(), file!(), line!(), $left, $right)));
        }
    };
    ($left:expr, $right:expr, $($arg:tt)*) => {
        if $left != $right {
            return Err(TestError(format!("Assertion failed in {} ({}:{}): (left: `{}`, right: `{}`): {}", std::any::type_name::<fn()>(), file!(), line!(), $left, $right, format_args!($($arg)*))));
        }
    };
}

#[macro_export]
macro_rules! assert_ne {
    ($left:expr, $right:expr) => {
        if $left == $right {
            return Err(TestError(format!("Assertion failed in {} ({}:{}): (left: `{}`, right: `{}`)", std::any::type_name::<fn()>(), file!(), line!(), $left, $right)));
        }
    };
    ($left:expr, $right:expr, $($arg:tt)*) => {
        if $left == $right {
            return Err(TestError(format!("Assertion failed in {} ({}:{}): (left: `{}`, right: `{}`): {}", std::any::type_name::<fn()>(), file!(), line!(), $left, $right, format_args!($($arg)*))));
        }
    };
}

#[macro_export]
macro_rules! fixup {
    ($cond:expr) => {
        if $cond {
            return Err(TestError(format!("Known issue fixed in {} ({}:{})", std::any::type_name::<fn()>(), file!(), line!())));
        } else {
            eprintln!("Warning: Known issue still present in {} ({}:{})", std::any::type_name::<fn()>(), file!(), line!());
        }
    };
    ($cond:expr, $($arg:tt)*) => {
        if $cond {
            return Err(TestError(format!("Known issue fixed in {} ({}:{}): {}", std::any::type_name::<fn()>(), file!(), line!(), format_args!($($arg)*))));
        } else {
            eprintln!("Warning: Known issue still present in {} ({}:{}): {}", std::any::type_name::<fn()>(), file!(), line!(), format_args!($($arg)*));
        }
    };
}

#[macro_export]
macro_rules! ignore {
    ($cond:expr) => {
        if $cond {
            eprintln!("Warning: Known issue fixed in {} ({}:{})", std::any::type_name::<fn()>(), file!(), line!());
        } else {
            eprintln!("Warning: Known issue still present in {} ({}:{})", std::any::type_name::<fn()>(), file!(), line!());
        }
    };
    ($cond:expr, $($arg:tt)*) => {
        if $cond {
            eprintln!("Warning: Known issue fixed in {} ({}:{})", std::any::type_name::<fn()>(), file!(), line!());
        } else {
            eprintln!("Warning: Known issue still present in {} ({}:{}): {}", std::any::type_name::<fn()>(), file!(), line!(), format_args!($($arg)*));
        }
    };
}

/// Holds a `String` to run `syd`.
/// Honours CARGO_BIN_EXE_syd environment variable.
pub static SYD: Lazy<String> =
    Lazy::new(|| env::var("CARGO_BIN_EXE_syd").unwrap_or("syd".to_string()));

/// Holds a `String` to run `syd-test-do`.
/// Honours CARGO_BIN_EXE_syd-test-do environment variable.
pub static SYD_DO: Lazy<String> =
    Lazy::new(|| env::var("CARGO_BIN_EXE_syd-test-do").unwrap_or("syd-test-do".to_string()));

/// Returns a `Command` to run `syd`.
/// Honours CARGO_BIN_EXE_syd environment variable.
pub fn syd() -> Command {
    static USE_VALGRIND: Lazy<bool> = Lazy::new(|| env::var("SYD_TEST_VALGRIND").ok().is_some());
    let mut cmd = Command::new(if *USE_VALGRIND { "valgrind" } else { "timeout" });
    if *USE_VALGRIND {
        cmd.arg("--leak-check=yes");
        cmd.arg("--track-origins=yes");
        cmd.arg("--track-fds=yes");
        cmd.arg("--trace-children=no");
        cmd.arg("--");
        cmd.arg("timeout");
    }
    cmd.arg("-sKILL");
    cmd.arg("15m");
    cmd.arg(&*SYD);
    cmd.env("RUST_BACKTRACE", "full");
    cmd.env("SYD_LOG", "trace");
    cmd.env("SYD_NO_SYSLOG", "1");
    cmd
}

/// Resembles the `which` command, finds a program in PATH.
pub fn which(command: &str) -> Result<String, Errno> {
    let out = Command::new("which")
        .arg(command)
        .output()
        .expect("execute which")
        .stdout;
    if out.is_empty() {
        return Err(Errno::ENOENT);
    }

    let bin = String::from_utf8_lossy(&out);
    let bin = bin.trim();
    Ok(canonicalize(bin)
        .map_err(|_| Errno::last())?
        .to_string_lossy()
        .into_owned())
}

/// Format a `Duration` into a human readable `String`.
pub fn format_duration(d: Duration) -> String {
    let total_seconds = d.as_secs();
    let hours = total_seconds / 3600;
    let minutes = (total_seconds % 3600) / 60;
    let seconds = total_seconds % 60;

    format!("{}h {}m {}s", hours, minutes, seconds)
}
